from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...repositories.project.project import (
    FormStage,
    JobStage,
    NotificationTrigger,
    ProjectRepository,
    ScriptStage,
    WorkflowTransition,
)
from .workflows import get_path


class WorkflowTest(TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_empty_project(self):
        path = get_path(3)
        self.assertEqual(path, [])

    def test_single_form(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.forms.append(form)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form"])

    def test_more_then_one_stage(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script",
                    type="finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.forms.append(form)
        project.scripts.append(script)
        ProjectRepository.save(project)

        path = get_path(2)
        self.assertEqual(path, ["form", "script"])

    def test_more_stages_then_limit(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script",
                    type="finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.forms.append(form)
        project.scripts.append(script)
        ProjectRepository.save(project)

        path = get_path(1)
        self.assertEqual(path, ["form"])

    def test_more_then_one_path(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="script",
                    target_id="script3",
                    type="finished",
                )
            ],
        )

        script3 = ScriptStage(
            id="script3",
            file="script3.py",
            title="Script3",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.scripts.append(script3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form0", "script2", "script3"])

    def test_more_than_one_initial_stage(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.forms.append(form1)
        project.forms.append(form2)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form1", "form2", "script2"])

    def test_disjointed_graphs(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="form",
                    target_id="form3",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form3 = FormStage(
            id="form3",
            path="form3",
            title="Form3",
            file="form3.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.forms.append(form1)
        project.forms.append(form2)
        project.forms.append(form3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form1", "form2", "form3"])

    def test_simple_loop(self):
        project = ProjectRepository.load()
        job0 = JobStage(
            id="job0",
            title="Job0",
            file="job0.py",
            workflow_position=(0, 0),
            schedule="* * * * *",
            workflow_transitions=[
                WorkflowTransition(
                    id="0",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )

        project.jobs.append(job0)
        project.forms.append(form0)
        project.scripts.append(script1)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["job0", "form0", "script1"])

    def test_dont_enter_in_loops(self):
        project = ProjectRepository.load()
        job0 = JobStage(
            id="job0",
            title="Job0",
            file="job0.py",
            workflow_position=(0, 0),
            schedule="* * * * *",
            workflow_transitions=[
                WorkflowTransition(
                    id="0",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )

        project.jobs.append(job0)
        project.forms.append(form0)
        project.scripts.append(script1)
        ProjectRepository.save(project)

        path = get_path(10)
        self.assertEqual(path, ["job0", "form0", "script1"])

    def test_clips_the_largest_path(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="form",
                    target_id="form1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
                WorkflowTransition(
                    id="3",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="form",
                    target_id="form3",
                    type="finished",
                ),
                WorkflowTransition(
                    id="5",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form3 = FormStage(
            id="form3",
            path="form3",
            title="Form3",
            file="form3.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="7",
                    target_type="script",
                    target_id="script3",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script3 = ScriptStage(
            id="script3",
            file="script3.py",
            title="Script3",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        project.forms.append(form0)
        project.forms.append(form1)
        project.scripts.append(script1)
        project.forms.append(form2)
        project.scripts.append(script2)
        project.forms.append(form3)
        project.scripts.append(script3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form0", "form1", "form2"])



from typing import List

import requests

from ..contracts_generated import (
    CommonRole,
)
from ..credentials import (
    resolve_headers,
)
from ..environment import CLOUD_API_ENDPOINT


class RolesRepository:
    @property
    def headers(self):
        return resolve_headers()

    def get_roles(self) -> List[CommonRole]:
        url = f"{CLOUD_API_ENDPOINT}/cli/roles"
        response = requests.get(url, headers=self.headers)
        return [CommonRole.from_dict(role) for role in response.json()]


def roles_repository_factory() -> RolesRepository:
    return RolesRepository()



import pathlib

from .environment import FILES_FOLDER
from .settings import Settings
from .utils.dot_abstra import PERSISTENT_DIR


def get_persistent_dir() -> pathlib.Path:
    if FILES_FOLDER:
        return pathlib.Path(FILES_FOLDER)

    dir = Settings.root_path / PERSISTENT_DIR
    dir.mkdir(parents=True, exist_ok=True)
    return dir



import datetime

from gunicorn.arbiter import Arbiter
from gunicorn.workers.base import Worker

from ..environment import set_SERVER_UUID, set_WORKER_UUID
from ..repositories import (
    execution_logs_repository,
    execution_repository,
    stage_run_repository,
)
from ..repositories.execution_logs import StdioLogEntry
from ..threaded import threaded
from ..utils import get_internal_id


@threaded
def async_child_exit(server: Arbiter, worker: Worker):
    app_id = get_internal_id(server, ensure=False)
    worker_id = get_internal_id(worker, ensure=False)

    killed_executions = execution_repository.find_for_worker(
        worker_id=worker_id,
        status="running",
        app_id=app_id,
    )

    # update executions
    for execution in killed_executions:
        # Add log entry
        err_log = StdioLogEntry(
            execution_id=execution.id,
            event="stderr",
            payload={"text": "Aborted: Server reached its memory limit"},
            created_at=datetime.datetime.now(),
            sequence=999999,
        )
        execution_logs_repository.save(err_log)

        # Update execution status
        execution.status = "failed"
        execution_repository.update(execution)

        # Update stage run
        if execution.stage_run_id:
            stage_run_repository.change_state(execution.stage_run_id, "failed")


#### HOOKS


# Called just before the master process is initialized.
def on_starting(server: Arbiter):
    server_uuid = get_internal_id(server)
    set_SERVER_UUID(server_uuid)


# Called just before a worker is forked.
def pre_fork(server: Arbiter, worker: Worker):
    worker_uuid = get_internal_id(worker)
    set_WORKER_UUID(worker_uuid)


# Called just after a worker has been exited, in the master process.
def child_exit(server: Arbiter, worker: Worker):
    async_child_exit(server, worker)



from ..widget_base import Output


class ProgressOutput(Output):
    type = 'progress-output'

    def __init__(self, current: float=50, total: float=100, **kwargs):
        self.set_props(dict(current=current, total=total, **kwargs))

    def set_props(self, props):
        self.current = props.get('current', 0)
        self.total = props.get('total', 0)
        self.text = props.get('text', '')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'current': self.current, 'total': self.
            total, 'text': self.text, 'fullWidth': self.full_width}



from typing import List

from ...settings import Settings
from ...utils.file import files_from_directory
from ..linter import LinterFix, LinterIssue, LinterRule


class AddEnvToAbstraIgnore(LinterFix):
    label = "Add env to abstra ignore"

    def fix(self):
        abstraignore_file = Settings.root_path / ".abstraignore"
        with abstraignore_file.open("a") as file:
            file.write("\n.env")


class EnvInBundleFound(LinterIssue):
    def __init__(self) -> None:
        self.label = "You have not ignored the .env file"
        self.fixes = [AddEnvToAbstraIgnore()]


class EnvInBundle(LinterRule):
    label = "You can't add .env to the bundle"
    type = "security"

    def find_issues(self) -> List[LinterIssue]:
        env_file = Settings.root_path / ".env"

        bundle_files = files_from_directory(Settings.root_path)

        if env_file in bundle_files:
            return [EnvInBundleFound()]
        else:
            return []



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import requests

from ..contracts_generated import CommonUserRoles
from ..environment import SIDECAR_HEADERS, SIDECAR_URL


class MembersRepository(ABC):
    @abstractmethod
    def get_member_roles(self, email: str) -> CommonUserRoles:
        pass


class LocalMembersRepository(MembersRepository):
    data: Optional[Dict[str, CommonUserRoles]]

    def __init__(self, database: Optional[Dict[str, CommonUserRoles]] = None):
        self.data = database

    def get_member_roles(self, email: str) -> CommonUserRoles:
        if self.data is None:
            return ["workflow_viewer"]
        return self.data.get(email, [])


class ProductionMembersRepository(MembersRepository):
    headers: Dict[str, str]

    def __init__(self, headers) -> None:
        self.headers = headers

    @classmethod
    def _request(
        cls,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers: Dict[str, str] = SIDECAR_HEADERS
        r = requests.request(
            method=method,
            url=f"{SIDECAR_URL}/users{path}",
            headers=headers,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    def _get_user_by_email(self, email: str):
        r = self._request("GET", "/", params={"email": email}, raise_for_status=False)
        if r.status_code == 404:
            return None
        return r.json()

    def get_member_roles(self, email: str) -> CommonUserRoles:
        user = self._get_user_by_email(email)
        if user is None:
            return []
        return user.get("roles", [])


def members_repository_factory() -> MembersRepository:
    if SIDECAR_URL is None:
        return LocalMembersRepository()
    else:
        headers = SIDECAR_HEADERS
        return ProductionMembersRepository(headers)



from ..widget_base import Input
from io import BufferedReader, TextIOWrapper
from typing import Optional, Union, List
from ..apis import upload_file
from ..response_types import FileResponse
from pathlib import Path


class CameraInput(Input):
    type = 'camera-input'
    empty_value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.layout = props.get('layout', 'list')
        self.disabled = props.get('disabled', False)
        self.value = props.get('initial_value', self.empty_value)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'hint': self.hint, 'value': self.serialize_value(), 'required':
            self.required, 'fullWidth': self.full_width, 'layout': self.
            layout, 'disabled': self.disabled, 'errors': self.errors}

    @staticmethod
    def __get_file_uri(value: Union[FileResponse, str, BufferedReader,
        TextIOWrapper]) ->str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value))
            else:
                return value
        if isinstance(value, FileResponse):
            return value.url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ''

    def serialize_value(self):
        if self.value is None:
            return None
        return CameraInput.__get_file_uri(self.value)

    def parse_value(self, value: Optional[List[str]]) ->Optional[Union[
        FileResponse, List[FileResponse]]]:
        if value is None:
            return self.empty_value
        file_response = FileResponse(value)
        return super().parse_value(file_response)



import json

import flask

from ...execution.execution import RequestData
from ...execution.hook_execution import HookExecution
from ...execution.stage_run_manager import (
    AttachedStageRunManager,
    DetachedStageRunManager,
)
from ...repositories import (
    execution_logs_repository,
    execution_repository,
    stage_run_repository,
)
from ...repositories.execution_logs import get_logs_output
from ...repositories.project.project import ProjectRepository
from ...repositories.stage_run import stage_run_repository_factory
from ...usage import usage
from ...utils import is_it_true
from ..workflow_engine import workflow_engine
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_hooks", __name__)

    @bp.route("/<path:id>", methods=["GET"])
    @usage
    def _get_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)
        return hook.editor_dto

    @bp.route("/", methods=["GET"])
    @usage
    def _get_hooks():
        return [f.editor_dto for f in controller.get_hooks()]

    @bp.route("/", methods=["POST"])
    @usage
    def _create_hook():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        hook = controller.create_hook(title, file, workflow_position)
        return hook.editor_dto

    @bp.route("/<path:id>", methods=["PUT"])
    @usage
    def _update_hook(id: str):
        changes = flask.request.json
        if not changes:
            flask.abort(400)

        hook = controller.update_stage(id, changes)
        return hook.editor_dto if hook else None

    @bp.route("/<path:id>", methods=["DELETE"])
    @usage
    def _delete_hook(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_hook(id, remove_file)
        return {"success": True}

    @bp.route("/<path:id>/run", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    @usage
    def _run_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        project = ProjectRepository.load()
        is_initial = project.is_initial(hook)

        stage_run_manager = AttachedStageRunManager(stage_run_repository)

        execution = HookExecution(
            hook,
            is_initial,
            request_data,
            execution_repository=execution_repository,
            execution_logs_repository=execution_logs_repository,
            stage_run_manager=stage_run_manager,
        )

        execution.run()
        body, status, headers = execution.get_response()
        workflow_engine.handle_execution_end(execution)

        return {
            "body": body,
            "status": status,
            "headers": headers,
            "output": get_logs_output(execution_logs_repository.get(execution.id)),
        }

    @bp.route("/<path:id>/test", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    @usage
    def _test_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        data = json.loads(controller.read_test_data())

        tmp_stage_run_repository = stage_run_repository_factory()

        project = ProjectRepository.load()
        is_initial = project.is_initial(hook)

        stage_run_manager = DetachedStageRunManager(tmp_stage_run_repository, data=data)

        execution = HookExecution(
            hook,
            is_initial,
            request_data,
            execution_repository=execution_repository,
            execution_logs_repository=execution_logs_repository,
            stage_run_manager=stage_run_manager,
        )

        execution.run()
        body, status, headers = execution.get_response()

        return {
            "body": body,
            "status": status,
            "headers": headers,
            "output": get_logs_output(execution_logs_repository.get(execution.id)),
        }

    return bp



import ast
from pathlib import Path
from typing import Dict, List, Set, Tuple

from ...repositories.project.project import ProjectRepository
from ...utils.code import function_called_args
from ..linter import LinterIssue, LinterRule


class UnsetMatchFound(LinterIssue):
    def __init__(self, key: str, path: Path, lineno: int):
        self.label = f"Unset key found: '{key}' in {path}:{lineno}"
        self.fixes = []


class UnsetMatchThread(LinterRule):
    label: str = "Undefined data to be matched"
    type: str = "bug"

    def find_issues(self) -> List[LinterIssue]:
        project = ProjectRepository.load()
        match_threads: Dict[str, Set[Tuple[Path, int]]] = {}
        data_sets: Dict[str, Set[Tuple[Path, int]]] = {}
        for python_file in project.project_files:
            try:
                match_thread_calls = function_called_args(
                    python_file.read_text(encoding="utf-8"),
                    ["abstra", "workflows"],
                    "match_thread",
                )
                set_data_calls = function_called_args(
                    python_file.read_text(encoding="utf-8"),
                    ["abstra", "workflows"],
                    "set_data",
                )

                if match_thread_calls is not None:
                    for function_call in match_thread_calls:
                        for arg in function_call:
                            if isinstance(arg, ast.Dict):
                                for key in arg.keys:
                                    if isinstance(key, ast.Constant) and isinstance(
                                        key.value, str
                                    ):
                                        match_threads.setdefault(key.value, set()).add(
                                            (python_file, key.lineno)
                                        )

                if set_data_calls is not None:
                    for function_call in set_data_calls:
                        for arg in function_call:
                            if isinstance(arg, ast.Constant) and isinstance(
                                arg.value, str
                            ):
                                data_sets.setdefault(arg.value, set()).add(
                                    (python_file, arg.lineno)
                                )
            except SyntaxError:
                continue
            except FileNotFoundError:
                continue

        issues = []
        for key, match_path in match_threads.items():
            if key not in data_sets:
                for path, lineno in match_path:
                    issues.append(UnsetMatchFound(key, path, lineno))

        return issues



import datetime
import json
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import requests

from ..contracts_generated import CommonUser, CommonUserRoles
from ..environment import SIDECAR_HEADERS, SIDECAR_URL
from ..logger import AbstraLogger
from ..settings import Settings
from ..utils.dot_abstra import LOCAL_USERS_FILE


class UsersRepository(ABC):
    @abstractmethod
    def get_user_roles(self, email: str) -> CommonUserRoles:
        pass

    @abstractmethod
    def get_user(self, email: str) -> Optional[CommonUser]:
        pass

    @abstractmethod
    def list_users(self) -> List[CommonUser]:
        pass

    @abstractmethod
    def insert_user(self, email: str, roles: CommonUserRoles) -> bool:
        pass

    @abstractmethod
    def delete_user(self, email: str) -> None:
        pass

    @abstractmethod
    def update_user(self, email: str, roles: CommonUserRoles) -> None:
        pass


class LocalUsersRepository(UsersRepository):
    data: Dict[str, CommonUserRoles]
    sign_up_policy: List[str]
    is_loaded: bool

    def __init__(self, data: Dict[str, CommonUserRoles] = {}) -> None:
        self.data = data
        self.sign_up_policy: List[str] = []
        self.is_loaded = False

    def __load(self):
        self.is_loaded = True
        self.local_cache = Settings.root_path.joinpath(LOCAL_USERS_FILE)
        try:
            with open(self.local_cache, "r") as f:
                self.data = {**self.data, **json.load(f)}
        except Exception as e:
            AbstraLogger.capture_exception(e)

    def __dump(self):
        with open(self.local_cache, "w") as f:
            json.dump(self.data, f)

    def get_user_roles(self, email: str) -> CommonUserRoles:
        user = self.get_user(email)
        if user is None:
            return []
        return user.roles

    def get_user(self, email: str) -> Optional[CommonUser]:
        if not self.is_loaded:
            self.__load()

        if email not in self.data:
            return None

        roles = self.data.get(email, [])

        return CommonUser(
            id=email,
            email=email,
            roles=roles,
            project_id=email,
            created_at=datetime.datetime.now().isoformat(),
        )

    def list_users(self) -> List[CommonUser]:
        if not self.is_loaded:
            self.__load()

        return [
            CommonUser(
                id=email,
                email=email,
                roles=roles,
                project_id=email,
                created_at=datetime.datetime.now().isoformat(),
            )
            for email, roles in self.data.items()
        ]

    def insert_user(self, email: str, roles: CommonUserRoles) -> bool:
        if not self.is_loaded:
            self.__load()

        self.data[email] = roles
        self.__dump()
        return True

    def delete_user(self, email: str) -> None:
        if email in self.data:
            del self.data[email]
            self.__dump()

    def update_user(self, email: str, roles: CommonUserRoles) -> None:
        if email in self.data:
            self.data[email] = roles
            self.__dump()


class ProductionUsersRepository(UsersRepository):
    headers: Dict[str, str]

    def __init__(self, headers) -> None:
        self.headers = headers

    @classmethod
    def _request(
        cls,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers: Dict[str, str] = SIDECAR_HEADERS
        r = requests.request(
            method=method,
            url=f"{SIDECAR_URL}/users{path}",
            headers=headers,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    def get_user(self, email: str) -> Optional[CommonUser]:
        r = self._request("GET", "/", params={"email": email}, raise_for_status=False)
        if r.status_code == 404:
            return None
        return CommonUser.from_dict(r.json())

    def get_user_roles(self, email: str) -> CommonUserRoles:
        user = self.get_user(email)
        if user is None:
            return []
        return user.roles

    def insert_user(self, email: str, roles: CommonUserRoles) -> bool:
        r = self._request(
            "POST", "/", body={"email": email, "roles": roles}, raise_for_status=False
        )
        return r.status_code == 201

    ##this is not expected, but is needed to keep the interface
    def list_users(self) -> List[CommonUser]:
        return []

    def delete_user(self, email: str) -> None:
        pass

    def update_user(self, email: str, roles: CommonUserRoles) -> None:
        pass


def users_repository_factory() -> UsersRepository:
    if SIDECAR_URL is None:
        return LocalUsersRepository()
    else:
        headers = SIDECAR_HEADERS
        return ProductionUsersRepository(headers)



import json
import pathlib

abstraignore = (pathlib.Path(__file__).parent / "abstraignore").read_text(
    encoding="utf-8"
)
new_form_code = (pathlib.Path(__file__).parent / "new_form.py").read_text(
    encoding="utf-8"
)
new_hook_code = (pathlib.Path(__file__).parent / "new_hook.py").read_text(
    encoding="utf-8"
)
new_job_code = (pathlib.Path(__file__).parent / "new_job.py").read_text(
    encoding="utf-8"
)
new_script_code = (pathlib.Path(__file__).parent / "new_script.py").read_text(
    encoding="utf-8"
)
vscode_launch_json = json.load(
    (pathlib.Path(__file__).parent / "vscode_launch.json").open()
)


def ensure_abstraignore(dir: pathlib.Path):
    path = dir / ".abstraignore"
    if not path.exists():
        path.write_text(abstraignore, encoding="utf-8")



import os
import unittest

from tests.fixtures import clear_dir, init_dir

from .credentials import (
    delete_credentials,
    get_credentials,
    resolve_headers,
    set_credentials,
)
from .utils.dot_abstra import CREDENTIALS_FILE, DOT_ABSTRA_FOLDER_NAME


class CredentialsTest(unittest.TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self):
        clear_dir(self.root)
        if "ABSTRA_API_TOKEN" in os.environ:
            del os.environ["ABSTRA_API_TOKEN"]

    def test_get_credentials_when_empty(self):
        self.assertIsNone(get_credentials())

    def test_get_credentials_when_file(self):
        self.root.joinpath(DOT_ABSTRA_FOLDER_NAME).mkdir(exist_ok=True)
        self.root.joinpath(CREDENTIALS_FILE).write_text("test")
        self.assertEqual(get_credentials(), "test")

    def test_get_credentials_when_env(self):
        os.environ["ABSTRA_API_TOKEN"] = "test"
        self.assertEqual(get_credentials(), "test")

    def test_set_credentials(self):
        set_credentials("test")
        self.assertEqual(get_credentials(), "test")

    def test_delete_credentials(self):
        set_credentials("test")
        delete_credentials()
        self.assertIsNone(get_credentials())

    def test_resolve_headers_when_empty(self):
        self.assertIsNone(resolve_headers())

    def test_resolve_headers_with_credentials(self):
        set_credentials("test")
        self.assertEqual(resolve_headers(), {"Api-Authorization": "Bearer test"})



from pathlib import Path
from tempfile import gettempdir
from unittest import TestCase

from .settings import SettingsController


class SettingsTest(TestCase):
    def test_settings_controller_set_valid_root_path(self):
        controller = SettingsController()
        tmpdir = gettempdir()
        controller.set_root_path(tmpdir)
        self.assertEqual(controller.root_path, Path(tmpdir))

    def test_settings_controller_set_deep_valid_root_path(self):
        controller = SettingsController()
        tmpdir = gettempdir()
        controller.set_root_path(tmpdir + "/test")
        self.assertEqual(controller.root_path, Path(tmpdir + "/test"))

    def test_settings_controller_no_default_server_port(self):
        controller = SettingsController()
        self.assertRaises(Exception, lambda: controller.server_port)

    def test_settings_controller_set_str_server_port(self):
        controller = SettingsController()
        controller.set_server_port("3000")
        self.assertEqual(controller.server_port, 3000)

    def test_settings_controller_set_int_server_port(self):
        controller = SettingsController()
        controller.set_server_port(3000)
        self.assertEqual(controller.server_port, 3000)



from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...repositories.requirements import RequirementsRepository
from ...server.controller.main import MainController
from .duplicate_package_in_requirements import DuplicatePackagesInRequirements


class TestDuplicatePackagesInRequirements(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_default_valid(self):
        MainController()
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

    def test_valid_with_no_requirements_file(self):
        MainController()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.unlink()
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

    def test_valid_with_no_duplicates(self):
        MainController()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==1.0.0\nflask==1.0.0")
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

    def test_invalid_with_equal_duplicates(self):
        MainController()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==1.0.0\nabstra==1.0.0")
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 1)
        self.assertEqual(len(issues[0].fixes), 1)

        issues[0].fixes[0].fix()

        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

    def test_invalid_with_distinct_versions_choose_first(self):
        MainController()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==1.0.0\nabstra==2.0.0")
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 1)
        self.assertEqual(len(issues[0].fixes), 2)

        issues[0].fixes[1].fix()

        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

        requirements = RequirementsRepository.load()
        self.assertEqual(len(requirements.libraries), 1)
        self.assertEqual(requirements.libraries[0].name, "abstra")
        self.assertEqual(requirements.libraries[0].version, "1.0.0")

    def test_invalid_with_distinct_versions_choose_second(self):
        MainController()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==1.0.0\nabstra==2.0.0")
        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 1)
        self.assertEqual(len(issues[0].fixes), 2)

        issues[0].fixes[0].fix()

        issues = DuplicatePackagesInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

        requirements = RequirementsRepository.load()
        self.assertEqual(len(requirements.libraries), 1)
        self.assertEqual(requirements.libraries[0].name, "abstra")
        self.assertEqual(requirements.libraries[0].version, "2.0.0")



import flask

from ...settings import Settings
from ..utils import send_from_dist
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_assets", __name__)

    @bp.get("/<path:path>")
    def _asset(path):
        return send_from_dist(path, dist_folder=Settings.root_path)

    return bp



from os.path import getmtime
from typing import List

from ...settings import Settings
from ..linter import LinterIssue, LinterRule


class EnvFileChangeFound(LinterIssue):
    def __init__(self) -> None:
        self.label = "Your .env file was updated. Please restart your development server to ensure consistency."
        self.fixes = []


class EnvFileChanged(LinterRule):
    label = "Changes to .env file detected"
    type = "warning"
    initial_last_modified = None
    initial_file_exists = None

    def find_issues(self) -> List[LinterIssue]:
        env_file = Settings.root_path / ".env"
        current_file_exists = env_file.exists()

        if EnvFileChanged.initial_file_exists is None:
            EnvFileChanged.initial_file_exists = current_file_exists
            if current_file_exists:
                EnvFileChanged.initial_last_modified = getmtime(env_file)
                return []

        if EnvFileChanged.initial_file_exists != current_file_exists:
            return [EnvFileChangeFound()]

        if (
            EnvFileChanged.initial_last_modified
            and EnvFileChanged.initial_last_modified < getmtime(env_file)
        ):
            return [EnvFileChangeFound()]

        return []
